/*
 * the algorithm is stolen from Cilk benchmark (blockedmul.cilk).
 * rewritten for MTCAMP for demonstraton purposes
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <st_mm.h>

#define BLOCK_EDGE 16

typedef struct block
{
  double b[BLOCK_EDGE][BLOCK_EDGE];
} * block_t;


#ifdef VISUALIZE
/* to avoid false sharing */
typedef struct slong {
  long x;
  long dummy[31];
} * slong_t;

static slong_t picture;
static int picture_size = 0;

static void alloc_picture(int n)
{
  picture_size = n / BLOCK_EDGE;
  picture = (slong_t)MM_ALLOC(picture_size * picture_size * sizeof(struct slong));
}

static void set_picture(int i, int j, int wid)
{
  assert(picture_size != 0);
  assert(picture_size > i);
  assert(picture_size > j);
  picture[i * picture_size + j].x = wid;
}

typedef unsigned char uchar;
typedef struct color
{
  uchar r, g, b;
} * color_t;

color_t make_color_table(int nh)
{
  int k = 2;
  uchar r, g, b;
  uchar p;
  color_t ct;
  while (k * k * k < nh) k++;
  ct = (color_t)MM_ALLOC(sizeof(struct color) * k * k * k);
  p = 255 / (k - 1);

  for (b = 0; b < k; b++) {
    for (g = 0; g < k; g++) {
      for (r = 0; r < k; r++) {
	int idx = b * k * k + g * k + r;
	if (idx < k * k * k) {
	  ct[idx].r = r * p;
	  ct[idx].g = g * p;
	  ct[idx].b = b * p;
	  idx++;
	} else {
	  fprintf(stderr, "wrong: idx = %d >= %dx%dx%d = %d\n", 
		  idx, k, k, k, k * k * k);
	  st_wg_die((void *)1);
	}
      }
    }
  }
  return ct;
}

static void output_picture(void)
{
  int i, j, nh;
  color_t ct;
  nh = st_n_current_workers();
  ct = make_color_table(nh);
  printf("P3\n");
  printf("%d %d\n", picture_size, picture_size);
  printf("255\n");
  for (i = 0; i < picture_size; i++) {
    for (j = 0; j < picture_size; j++) {
      int w = picture[i * picture_size + j].x;
      printf("%d %d %d\n", ct[w].r, ct[w].g, ct[w].b);
    }
  }
}
#endif /* VISUALIZE */


void serial_mul(block_t a, block_t b, block_t r)
{
  int i, j, k;
  for (i = 0; i < BLOCK_EDGE; i++) {
    for (j = 0; j < BLOCK_EDGE; j++) {
      double s = 0.0;
      for (k = 0; k < BLOCK_EDGE; k++) {
	s += a->b[i][k] * b->b[k][j];
      }
      r->b[i][j] = s;
    }
  }
}

void serial_add(block_t t, block_t r)
{
  int i, j;
  for (i = 0; i < BLOCK_EDGE; i++) {
    for (j = 0; j < BLOCK_EDGE; j++) {
      r->b[i][j] += t->b[i][j];
    }
  }
}

void mat_add(int n, block_t t, block_t r)
{
  if (n == 1) {
    serial_add(t, r);
  } else {
    mtc_sync {
      int halfn;
      ST_POLLING();
      assert(n % 2 == 0);
      halfn = n / 2;
      mtc_fork mat_add(halfn, t, r);
      mat_add(halfn, t + halfn, r + halfn);
    }
  }
}

/*
 * Blocked matrix multiply is done as follows:
 *
 *      A      x      B      =          R
 *
 *  | C | D |     | G | H |     | CG+DI | CH+DJ |
 *  |---+---|  x  |---+---|  =  |-------+-------|
 *  | E | F |     | I | J |     | EG+FI | EH+FJ |
 *
 */
/* n = the number of blocks in a row (or in a column) */
void mat_mul(block_t a, block_t b, int n, block_t r, int rowid, int colid)
{
  if (n == 1) {
#ifdef VISUALIZE
    set_picture(rowid, colid, st_worker_id());
#endif /* VISUALIZE */
    serial_mul(a, b, r);
  } else {
    int halfn = n / 2;
    int halfn2 = halfn * halfn;
    block_t c, d, e, f, g, h, i, j;
    block_t cg, ch, eh, eg, di, dj, fj, fi;
    block_t tmp;
    assert(n % 2 == 0);
    ST_POLLING();

    tmp = (block_t)MM_ALLOC(n * n * sizeof(struct block));

    c = a + 0 * halfn2;
    d = a + 1 * halfn2;
    e = a + 2 * halfn2;
    f = a + 3 * halfn2;

    g = b + 0 * halfn2;
    h = b + 1 * halfn2;
    i = b + 2 * halfn2;
    j = b + 3 * halfn2;

    cg = r + 0 * halfn2;
    ch = r + 1 * halfn2;
    eh = r + 2 * halfn2;
    eg = r + 3 * halfn2;
    
    di = tmp + 0 * halfn2;
    dj = tmp + 1 * halfn2;
    fj = tmp + 2 * halfn2;
    fi = tmp + 3 * halfn2;
    
#if 0				/* serial */
    mat_mul(c, g, halfn, cg, rowid,         colid);
    mat_mul(d, i, halfn, di, rowid,         colid);
    mat_mul(c, h, halfn, ch, rowid,         colid + halfn);
    mat_mul(d, j, halfn, dj, rowid,         colid + halfn);
    mat_mul(e, h, halfn, eh, rowid + halfn, colid + halfn);
    mat_mul(f, j, halfn, fj, rowid + halfn, colid + halfn);
    mat_mul(e, g, halfn, eg, rowid + halfn, colid);
    mat_mul(f, i, halfn, fi, rowid + halfn, colid);
#else  /* parallel */
    mtc_sync {
      ST_POLLING();
      mtc_fork {
	mtc_fork {
	  mat_mul(c, g, halfn, cg, rowid,         colid);
	  mat_mul(d, i, halfn, di, rowid,         colid);
	}
	ST_POLLING();
	mat_mul(c, h, halfn, ch, rowid,         colid + halfn);
	mat_mul(d, j, halfn, dj, rowid,         colid + halfn);
      }
      ST_POLLING();
      mtc_fork {
	mat_mul(e, h, halfn, eh, rowid + halfn, colid + halfn);
	mat_mul(f, j, halfn, fj, rowid + halfn, colid + halfn);
      }
      ST_POLLING();
      mat_mul(e, g, halfn, eg, rowid + halfn, colid);
      mat_mul(f, i, halfn, fi, rowid + halfn, colid);
    }
#endif

    mat_add(n * n, tmp, r);
  }
}

void mat_init(block_t a, int n)
{
  if (n == 1) {
    int i, j;
    for (i = 0; i < BLOCK_EDGE; i++) {
      for (j = 0; j < BLOCK_EDGE; j++) {
	a->b[i][j] = 1.0;
      }
    }
  } else {
    mtc_sync {
      int halfn = n / 2;
      assert(n % 2 == 0);
      mtc_fork mat_init(a, halfn);
      mat_init(a + halfn, halfn);
    }
  }
}

void check_result(block_t r, double val, int n)
{
  if (n == 1) {
    int i, j;
    for (i = 0; i < BLOCK_EDGE; i++) {
      for (j = 0; j < BLOCK_EDGE; j++) {
	if (r->b[i][j] != val) {
	  printf("bad result (%4d, %4d) = %8f != %8f\n",
		 i, j, r->b[i][j], val);
	  return;
	}
      }
    }
  } else {
    mtc_sync {
      int halfn = n / 2;
      assert(n % 2 == 0);
      mtc_fork check_result(r, val, halfn);
      check_result(r + halfn, val, halfn);
    }
  }
}

int main(int argc, char **argv)
{
  int n, nblocks;
  block_t a, b, r;
  long t0, t1;
  if (argc != 2) {
    printf("Usage: %s size\n", argv[0]);
    return -1;
  }
  n = atoi(argv[1]);
  fprintf(stderr, "%d x %d blocked matrix multiply\n", n, n);
  if (((n & (n - 1)) != 0) || ((n % BLOCK_EDGE) != 0)) {
    printf("%d: matrix size must be a power of 2"
	   " and a multiple of %d\n", n, BLOCK_EDGE);
    return -1;
  }
  nblocks = (n / BLOCK_EDGE) * (n / BLOCK_EDGE);

  a = (block_t)MM_ALLOC(nblocks * sizeof(struct block));
  b = (block_t)MM_ALLOC(nblocks * sizeof(struct block));
  r = (block_t)MM_ALLOC(nblocks * sizeof(struct block));
#ifdef VISUALIZE
  alloc_picture(n);
#endif /* VISUALIZE */
  mtc_sync {
    mtc_fork mat_init(a, nblocks);
    mat_init(b, nblocks);
  }
  t0 = st_current_time_ms();
  mat_mul(a, b, n / BLOCK_EDGE, r, 0, 0);
  t1 = st_current_time_ms();
  fprintf(stderr, "%d msec\n", t1 - t0);
  check_result(r, n, nblocks);
#ifdef VISUALIZE
  output_picture();
#endif /* VISUALIZE */
  return 0;
}

