/*
 * grammar.cc
 */

#include "cfg.h"

void grammar::set_binary_rule(int mother, int left, int right, 
			      int head, st_region_t rg)
{
  binary_rules[left][right] 
    = cons(((head << SYMBOL_TABLE_MOTHER_WIDTH) | mother), 
	   binary_rules[left][right], rg);
}

void grammar::set_unary_rule(int mother, int child, st_region_t rg)
{
  unary_rules[child] 
    = cons(((UNARY_HEAD << SYMBOL_TABLE_MOTHER_WIDTH) | mother), 
	   unary_rules[child], rg);
}

void grammar::set_lex_rule(int mother, int child, st_region_t rg)
{
  unary_rules[child] 
    = cons(((LEX_HEAD << SYMBOL_TABLE_MOTHER_WIDTH) | mother), 
	   unary_rules[child], rg);
}

#define MAX_FILENAMES 256
void grammar::read_grammar(char * prefix, st_region_t rg)
{
  char filename[MAX_FILENAMES];
  sprintf(filename, "%s.bin.grammar", prefix);
  FILE * s = fopen(filename, "r");
  if (s == 0) {
    fprintf(stderr, "can't open %s\n", filename);
    perror("read_grammar");
    st_wg_exit((void *)1);
  }
  int i = 0;
  while (!feof(s)) {
    int mother = read_token_int(s, CHAR_SPACE);
    int left = read_token_int(s, CHAR_SPACE);
    int right = read_token_int(s, CHAR_SPACE);
    int head = read_token_int(s, CHAR_NEWLINE);
    set_binary_rule(mother, left, right, head, rg);
    i++;
  }
  fclose(s);
  
  sprintf(filename, "%s.una.grammar", prefix);
  s = fopen(filename, "r");
  if (s == 0) {
    fprintf(stderr, "can't open %s\n", filename);
    perror("read_grammar");
    st_wg_exit((void *)1);
  }
  i = 0;
  while (!feof(s)) {
    int mother = read_token_int(s, CHAR_SPACE);
    int child = read_token_int(s, CHAR_NEWLINE);
    set_unary_rule(mother, child, rg);
    i++;
  }
  fclose(s);
  
  sprintf(filename, "%s.lex.grammar", prefix);
  s = fopen(filename, "r");
  if (s == 0) {
    fprintf(stderr, "can't open %s\n", filename);
    perror("read_grammar");
    st_wg_exit((void *)1);
  }
  i = 0;
  while (!feof(s)) {
    int mother = read_token_int(s, CHAR_SPACE);
    int child = read_token_int(s, CHAR_NEWLINE);
    set_lex_rule(mother, child, rg);
  }
  fclose(s);
}

void grammar::feed_grammar(char * grammar_file, st_region_t rg)
{
  st_list_t(int) ** b = new (rg) st_list_t(int)* [bin_rule_size];

  int i, j;
  for (i = 0; i < bin_rule_size; i++) {
    b[i] = new (rg) st_list_t(int) [bin_rule_size];

    for (j = 0; j < bin_rule_size; j++) {
      b[i][j] = 0;
    }
  }
  binary_rules = b;
    
  st_list_t(int) * u = new (rg) st_list_t(int) [una_rule_size];
  for (i = 0; i < una_rule_size; i++)	{
    u[i] = 0;
  }
  unary_rules = u;

  read_grammar(grammar_file, rg);
}

grammar * make_grammar(char * grammar_file, st_region_t g_rg)
{
  grammar * grm = new (g_rg) grammar();
  grm->feed_grammar(grammar_file, g_rg);
  return grm;
}

static inline st_list_t(edge_t) 
cons_edge(int head, int s0, int k, int s1, st_list_t(edge) d, st_region_t rg)
{
  st_list_t(edge_t) 
    ce = (st_list_t(edge_t))MM_REGION_MALLOC(sizeof(st_list<edge_t>), rg);
  ce->car.k_head = MK_EDGE_K_HEAD(k, head);
  ce->car.s0_s1 = MK_EDGE_S0_S1(s0, s1);
  ce->cdr = d;
  return ce;
}

/* merge NEW_SYMS and ORIG_SYMS.  return the merged list.
   do not add symbols which are already found in ORIG_SYMS.  judges if
   a symbol is in ORIG_SYMS by array GEN_EDGES, which is an array indexed
   by a symbol. GEN_EDGES[s] is zero is symbols S has not been generated. */

static inline st_list_t(int) 
add_symbols(st_list_t(int) new_syms, st_list_t(int) orig_syms,
	    st_list_t(edge_t) * gen_edges, int s0, int k, int s1,
	    st_region_t rg)
{
  for (st_list_t(int) p = new_syms; p; p = p->cdr) {
    int m = p->car;
    int mother = (m & SYMBOL_TABLE_MOTHER_MASK);
    int head = (m >> SYMBOL_TABLE_MOTHER_WIDTH);
    st_list_t(int) osp;
    /* gen_edges need no locks */
    gen_edges[0] = cons_edge(head, s0, k, s1, gen_edges[0], rg);
    /* linear search */
    for (osp = orig_syms; osp; osp = osp->cdr)
      if (osp->car == mother) break;
    if (osp == 0) orig_syms = cons(mother, orig_syms, rg);
  }
  return orig_syms;
}

/* for each element s0 in S0
     for each element s1 in S1
         find all rules in grm of the form s -> s0 s1
         and accumulate all S on SYMBOLS */

st_list_t(int) 
grammar::apply_binary_rules(st_list_t(int) S0, st_list_t(int) S1, int k, 
			    st_list_t(int) symbols, 
			    st_list_t(edge_t) * gen_edges, 
			    st_region_t rg)
{
  st_list_t(int) ** b = binary_rules;
  for (st_list_t(int) p0 = S0; p0; p0 = p0->cdr) {
    int s0 = p0->car;
    for (st_list_t(int) p1 = S1; p1; p1 = p1->cdr) {
      int s1 = p1->car;
      st_list_t(int) new_syms = b[s0][s1];
      symbols = add_symbols(new_syms, symbols, gen_edges, s0, k, s1, rg);
    }
  }
  return symbols;
}


/* rappend (x - ox, y); */
template<class T>
static inline st_list_t(T) 
diff_rappend(st_list_t(T) x, st_list_t(T) ox, st_list_t(T) y, st_region_t rg)
{
  st_list_t(T) p;
  for (p = x; p != ox; p = p->cdr) {
    y = cons(p->car, y, rg);
  }
  return y;
}

/* for each element i in I
     find symbols which are transitively generated by applying 
     unary rules from symbol i.
     accumulate them on SYMBOLS. */

st_list_t(int) 
grammar::apply_unary_rules(st_list_t(int) I, st_list_t(int) symbols, 
			   st_list_t(edge_t) * gen_edges, st_region_t rg)
{
  st_list_t(int) * u = unary_rules;
  st_list_t(int) ip = I;
  while (ip) {
    int i = ip->car;
    st_list_t(int) new_syms0 = u[i];
    st_list_t(int) new_syms 
      = add_symbols(new_syms0, symbols, gen_edges, i, i, i + 1, rg);
    ip = diff_rappend(new_syms, symbols, ip->cdr, rg);
    symbols = new_syms;
  }
  return symbols;
}

