/*
 * force.cc
 */

#include "def.h"

#if COUNT_NEWTONS
long n_newton = 0;
#endif

inline vect_t 
newton_acceleration (t_real mass, vect_t cg, vect_t pos, st_region_t rg)
{
#if COUNT_NEWTONS
  st_fetch_and_add_long(&n_newton, 1);
#endif
  t_real x0 = VX(pos), y0 = VY(pos), z0 = VZ(pos);
  t_real x1 = VX(cg), y1 = VY(cg), z1 = VZ(cg);
  t_real dx = x1 - x0, dy = y1 - y0, dz = z1 - z0;
  t_real l2 = dx * dx + dy * dy + dz * dz;
  t_real k = mass / (l2 * rsqrt (l2));
  return make_vect(k * dx, k * dy, k * dz, rg);
}    

vect_t space::calc_accel(vect_t pos, st_region_t rg)
{
  switch (state) {
  case ONE_PARTICLE: 
    {
      return make_vect(0.0, 0.0, 0.0, rg);
    }
  case MULTIPLE_PARTICLES: 
    {
      int idx = select_covering_rectangle(pos, area);
      t_real resultx = 0.0, resulty = 0.0, resultz = 0.0;
      for (int i = 0; i < N_CHILDREN; i++) {
	space * s = subspaces[i];
	if (i == idx) {
          st_assert(s);
	  vect_t sv = s->calc_accel(pos, rg);
	  resultx += VX(sv);
	  resulty += VY(sv);
	  resultz += VZ(sv);
	} else {
	  if (s) {
	    vect_t sv = s->calc_accel1(pos, rg);
	    resultx += VX(sv);
	    resulty += VY(sv);
	    resultz += VZ(sv);
	  }
	}
      }
      return make_vect(resultx, resulty, resultz, rg);
    }
  default: 
    {
      printf("NO_PARTICLES!\n"); 
      st_wg_die((void *)1);
      return make_vect(0.0, 0.0, 0.0, rg); /* never reach */
    }
  }
}

#if 0
const t_real approx_theta = 0.67 * 0.67; /* 1.05263 */
#else
const t_real approx_theta = 1.0 * 1.0; /* 1.05263 */
#endif

vect_t space::calc_accel1(vect_t pos, st_region_t rg)
{
  switch (state) {
  case ONE_PARTICLE: 
    {
      return newton_acceleration(mass, cg, pos, rg);
    }
  case MULTIPLE_PARTICLES: 
    {
      t_real x0 = VX(pos), y0 = VY(pos), z0 = VZ(pos);
      t_real x1 = VX(cg), y1 = VY(cg), z1 = VZ(cg);
      t_real dx = x1 - x0, dy = y1 - y0, dz = z1 - z0;
      t_real l2 = dx * dx + dy * dy + dz * dz;
      if (diameter2 < l2 * approx_theta) {
#if COUNT_NEWTONS
	st_fetch_and_add_long(&n_newton, 1);
#endif
	t_real k = mass / (l2 * rsqrt (l2));
	return make_vect(k * dx, k * dy, k * dz, rg);
      } else {
	t_real resultx = 0.0, resulty = 0.0, resultz = 0.0;
	for (int i = 0; i < N_CHILDREN; i++) {
	  space * s = subspaces[i];
	  if (s) {
	    vect_t sv = s->calc_accel1(pos, rg);
	    resultx += VX(sv);
	    resulty += VY(sv);
	    resultz += VZ(sv);
	  }
	}
	return make_vect(resultx, resulty, resultz, rg);
      }
    }
  default:
    {
      printf("NO_PARTICLES!\n"); 
      st_wg_die((void *)1);
      return make_vect(0.0, 0.0, 0.0, rg); /* never reach */
    }
  }
}

void particle::set_accel(space * sp, st_region_t rg)
{
  vect_t a = sp->calc_accel(pos, rg);
  accel = a;
}

#if BLOCK_TEST
int block_counter = 0;
int block_frequency = 0;
#endif
void set_accels_dac(particle ** particles, int begin_idx, int end_idx, 
		    space * sp, st_region_t rg
#if STHREADS && !MTCAMP
		    , st_join_counter_t * ack
#endif
		    )
{
  ST_POLLING();
#if BLOCK_TEST
  if (block_frequency) {
    block_counter--;
    if (block_counter == 0) {
      block_counter = block_frequency;
      st_yield();
    }
  }
#endif
  if (end_idx - begin_idx < 5) {
    int i;
    
    for (i = begin_idx; i <= end_idx; i++) {
      particles[i]->set_accel(sp, rg);
    }
  } else {
#if MTCAMP
    mtc_sync {
#elif STHREADS
      st_join_counter_t cack[1];
      st_join_counter_init(cack, 2);
#endif /* MTCAMP || STHREADS */
      
      int c = (begin_idx + end_idx - 1) / 2;
      
      /* fork left branch */
#if MTCAMP
      mtc_fork {
	set_accels_dac(particles, begin_idx, c, sp, rg);
      }
#elif STHREADS
      ST_THREAD_CREATE(set_accels_dac(particles, 
				      begin_idx, c, sp, rg, cack));
#else  /* SEQUENTIAL */
      set_accels_dac(particles, begin_idx, c, sp, rg);
#endif /* MTCAMP || STHREADS */
      
      /* call right branch */
      set_accels_dac(particles, c + 1, end_idx, sp, rg
#if STHREADS && !MTCAMP
		     , cack
#endif
		     );
      
      /* synchronize two recursive calls */
      
#if MTCAMP
      ---;
#elif STHREADS
      st_join_counter_wait(cack);
#endif
#if MTCAMP
    } /* mtc_sync */
#endif
  }
  
  /* say I have done */
#if STHREADS && !MTCAMP
  st_join_counter_finish(ack);
#endif
  ST_POLLING();
}

void set_accels(particle ** particles, int n_particles, 
		space * sp, st_region_t rg)
{
#if 0 
    int i;
#pragma MP taskloop maxcpus(64)
#pragma MP taskloop private(i)
#pragma MP taskloop readonly(particles, n_particles, sp)
    for (i = 0; i < n_particles; i++) {
      particles[i]->set_accel(sp);
    }
#elif MTCAMP
    set_accels_dac(particles, 0, n_particles - 1, sp, rg);
#elif STHREADS
    st_join_counter_t ack[1];
    st_join_counter_init(ack, 1);
    set_accels_dac(particles, 0, n_particles - 1, sp, rg, ack);
    st_join_counter_wait(ack);
#else  /* SEQUENTIAL */
    set_accels_dac(particles, 0, n_particles - 1, sp, rg);
#endif
}

void particle::move(t_real dt, st_region_t rg)
{
  vect_t vel_prime = v_plus_v (vel, k_times_v(dt, accel, rg), rg);
  pos = v_plus_v (pos, k_times_v(dt, vel_prime, rg), rg);
  vel = vel_prime;
}

void move_particles(particle ** particles, int n_particles, 
		    t_real dt, st_region_t rg)
{
  int i;
#if 0/*endo debug */
  for (i = 0; i < n_particles; i++) {
      particle *p = particles[i];
      vect_t vel_prime = v_plus_v (p->vel, k_times_v(dt, p->accel, rg), rg);
      p->pos = v_plus_v (p->pos, k_times_v(dt, vel_prime, rg), rg);
      p->vel = vel_prime;
      if (i < 10) {
	  printf("p[%d]: accel ", i);
	  print_vect(particles[i]->accel);
	  printf("\n");
	  printf("p[%d]: vel ", i);
	  print_vect(particles[i]->vel);
	  printf("\n");
	  printf("p[%d]: pos ", i);
	  print_vect(particles[i]->pos);
	  printf("\n");
      }
  }
#else
  for (i = 0; i < n_particles; i++) {
      particles[i]->move(dt, rg);
  }
#endif
}

