/*
 * Copyright (c) 1998-1999 by Toshio Endo. All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */

/* a sample program for the parallel garbage collection library. */
/* this program counts the leaves of binary tree several times. */
/* the gc library must be constructed as `pthread version'. */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/time.h>
#include "sgc.h"

/* #define EXPLICIT_FREE */
/* #define DONT_GC */

typedef struct node {
    int isleaf;
    struct node *left;
    struct node *right;
} node_t;

typedef struct task {
    node_t *node;
    struct task *prev;
    struct task *next;
} task_t;

struct global_env;

typedef struct local_env {
    pthread_t tid;
    task_t *task_stack_top;
    task_t *task_stack_bottom;
    pthread_mutex_t ml;
    int result;
    struct global_env *g_envp;
} local_env_t;

typedef struct global_env {
    int npe;
    local_env_t *larray;

    int busy_count;
    pthread_mutex_t ml;
} global_env_t;

node_t *make_tree(int depth, global_env_t *g_envp);
int push_task(local_env_t *envp, node_t *node);
node_t *pop_task(local_env_t *envp);
node_t *steal_task(local_env_t *envp);
void work_with_stack(local_env_t *my_envp);
int steal_or_terminate(local_env_t *my_envp);
int local_main_loop(local_env_t *my_envp);
void *local_func(void *arg);

node_t *make_tree(int depth, global_env_t *g_envp)
{
    node_t *node;
    node = (node_t *)GC_malloc(sizeof(node_t));
    if (depth == 0) {
	node->isleaf = 1;
    }
    else {
	node->isleaf = 0;
	node->left = make_tree(depth-1, g_envp);
	node->right = make_tree(depth-1, g_envp);
    }
    return node;
}

#ifdef EXPLICIT_FREE
void free_tree(node_t *node)
{
    if (node->isleaf) {
	GC_free(node);
    }
    else {
	free_tree(node->left);
	free_tree(node->right);
	GC_free(node);
    }
}
#endif

int push_task(local_env_t *envp, node_t *node)
{
    task_t *newtask = (task_t *)GC_malloc(sizeof(task_t));

    pthread_mutex_lock(&envp->ml);
    newtask->node = node;
    newtask->next = envp->task_stack_top;
    newtask->prev = NULL;
    if (envp->task_stack_top == NULL) {
	envp->task_stack_bottom = newtask;
    }
    else {
	envp->task_stack_top->prev = newtask;
    }
    envp->task_stack_top = newtask;
    pthread_mutex_unlock(&envp->ml);

    return 0;
}

node_t *pop_task(local_env_t *envp)
{
    task_t *task;
    node_t *node;

    if (envp->task_stack_top == NULL) {
	/* no task */
	return NULL;
    }

    pthread_mutex_lock(&envp->ml);
    
    if (envp->task_stack_top == NULL) {
	/* no task */
	pthread_mutex_unlock(&envp->ml);
	return NULL;
    }

    task = envp->task_stack_top;
    if (task->next == NULL) {
	envp->task_stack_bottom = NULL;
    }
    else {
	task->next->prev = NULL;
    }
    envp->task_stack_top = task->next;

    pthread_mutex_unlock(&envp->ml);
    node = task->node;
#ifdef EXPLICIT_FREE
    GC_free(task);
#endif
    return node;
}

node_t *steal_task(local_env_t *envp)
{
    task_t *task;
    node_t *node;

    if (envp->task_stack_top == NULL) {
	/* no task */
	return NULL;
    }

    pthread_mutex_lock(&envp->ml);
    
    if (envp->task_stack_top == NULL) {
	/* no task */
	pthread_mutex_unlock(&envp->ml);
	return NULL;
    }

    task = envp->task_stack_bottom;
    if (task->prev == NULL) {
	envp->task_stack_top = NULL;
    }
    else {
	task->prev->next = NULL;
    }
    envp->task_stack_bottom = task->prev;

    pthread_mutex_unlock(&envp->ml);
    node = task->node;
#ifdef EXPLICIT_FREE
    GC_free(task);
#endif
    return node;
}

void work_with_stack(local_env_t *my_envp)
{
    node_t *node;
    while ((node = pop_task(my_envp)) != NULL) {
	if (node->isleaf) {
	    /* count leaf */
	    my_envp->result ++;
	}
	else {
	    push_task(my_envp, node->left);
	    push_task(my_envp, node->right);
	}
    }
}

int steal_or_terminate(local_env_t *my_envp)
{
    global_env_t *g_envp = my_envp->g_envp;

    pthread_mutex_lock(&g_envp->ml);
    g_envp->busy_count--;
    if (g_envp->busy_count == 0) {
	pthread_mutex_unlock(&g_envp->ml);
	return 0;
    }
    pthread_mutex_unlock(&g_envp->ml);

    while (1) {
	int j;
	for (j = 0; j < g_envp->npe; j++) {
	    node_t *node;
	    pthread_mutex_lock(&g_envp->ml);
	    node = steal_task(&g_envp->larray[j]);
	    if (node) {
		g_envp->busy_count++;
		pthread_mutex_unlock(&g_envp->ml);
		push_task(my_envp, node);
		return 1;
	    }
	    pthread_mutex_unlock(&g_envp->ml);
	}
	pthread_mutex_lock(&g_envp->ml);
	if (g_envp->busy_count == 0) {
	    pthread_mutex_unlock(&g_envp->ml);
	    return 0;
	}
	pthread_mutex_unlock(&g_envp->ml);
    }
}

int local_main_loop(local_env_t *my_envp)
{
    while (1) {
	work_with_stack(my_envp);
	if (steal_or_terminate(my_envp) == 0) {
	    break;
	}
    }
    return 0;
}

void *local_func(void *arg)
{
    local_env_t *my_envp = (local_env_t *)arg;

    local_main_loop(my_envp);

    return NULL;
}

#ifdef SDSM_TEST
int S_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
    int nut;
    int ngct;
    int depth;
    int ntimes;
    int rcd;
    int i, j;
    local_env_t *larray;
    global_env_t *g_envp;
    struct timeval st, et;

    if (argc < 4) {
	fprintf(stderr,
		"usage: %s tree-depth #times #user-threads [#gc-threads]\n",
		argv[0]);
	fprintf(stderr, "example: %s 16 10 4 4\n", argv[0]);
	exit(1);
    }

    depth = atoi(argv[1]);
    if (depth > 30) {
	fprintf(stderr, "too big `depth'\n");
	exit(1);
    }
    ntimes = atoi(argv[2]);
    nut = atoi(argv[3]);
    if (argc >= 5) {
	ngct = atoi(argv[4]);
    }
    else {
	ngct = nut;
    }

    gettimeofday(&st, NULL);

    /* initialize gc library */
    rcd = SGC_init(ngct, NULL);
    if (rcd < 0) {
	fprintf(stderr, "Failed to initialize GC library.\n");
	fprintf(stderr, "This program uses pthread.\n");
	fprintf(stderr, "Configuration may be wrong in config.mk.\n");
	fprintf(stderr, "Confirm `SGC_THREAD_PKG=SGC_PTHREAD'\n");
	fprintf(stderr, "If configuration is ok, there are possibly bugs in SGC library...\n");
    }
#ifdef DONT_GC
    GC_dont_gc = 1;
#endif
    /* initialize global env */
    g_envp = (global_env_t *)GC_malloc(sizeof(global_env_t));
    g_envp->npe = nut;
    pthread_mutex_init(&g_envp->ml, NULL);

    /* initialize all local env */
    larray = (local_env_t *)GC_malloc(nut * sizeof(local_env_t));
    for (j = 0; j < nut; j++) {
	pthread_mutex_init(&larray[j].ml, NULL);
	larray[j].g_envp = g_envp;
    }

    g_envp->larray = larray;

    /* iteration */
    for (i = 0; i < ntimes; i++) {
	struct timeval t0, t1, t2;
	node_t *tree;
	pthread_attr_t pattr;
	int sumresult;

	for (j = 0; j < nut; j++) {
	    larray[j].task_stack_top = NULL;
	    larray[j].task_stack_bottom = NULL;
	    larray[j].result = 0;
	}
	g_envp->busy_count = nut;

	gettimeofday(&t0, NULL);

	/* make tree */
	tree = make_tree(depth, g_envp);
	gettimeofday(&t1, NULL);

	/* push root node */
	push_task(&larray[0], tree);

	/* create child threads */
	pthread_attr_init(&pattr);
	pthread_attr_setscope(&pattr, PTHREAD_SCOPE_SYSTEM);
	for (j = 1; j < nut; j++) {
	    GC_pthread_create(&larray[j].tid, &pattr,
			   local_func, (void *)&larray[j]);
	}
	/* work of main thread */
	local_main_loop(&larray[0]);
	/* join all child threads */
	for (j = 1; j < nut; j++) {
	    GC_pthread_join(larray[j].tid, NULL);
	}

	gettimeofday(&t2, NULL);

#ifdef EXPLICIT_FREE
	free_tree(tree);
	gettimeofday(&t3, NULL);
#endif

	/* sum result */
	sumresult = 0;
	for (j = 0; j < nut; j++) {
	    sumresult += larray[j].result;
	}
	printf("number of leaf is %d\n", sumresult);
	printf("It took %ld ms to make tree (1 thread)\n",
	       (t1.tv_sec-t0.tv_sec)*1000 + (t1.tv_usec-t0.tv_usec)/1000);
	printf("It took %ld ms to count leaf (%d thread)\n",
	       (t2.tv_sec-t1.tv_sec)*1000 + (t2.tv_usec-t1.tv_usec)/1000,
	       nut);
#ifdef EXPLICIT_FREE
	printf("It took %d ms to free tree (1 thread)\n",
	       (t3.tv_sec-t2.tv_sec)*1000 + (t3.tv_usec-t2.tv_usec)/1000);
#endif
    }

    printf("%ld collection (%d gc threads) occured.\n", GC_gc_no, ngct);

    gettimeofday(&et, NULL);
    printf("It took %ld ms in total\n",
	   (et.tv_sec-st.tv_sec)*1000 + (et.tv_usec-st.tv_usec)/1000);

    return 0;
}
